/* This file is part of Om.  Copyright (C) 2005 Dave Robillard.
 * 
 * Om is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * Om is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "ControlGroups.h"
#include "PortModel.h"
#include "OmGtk.h"
#include "Controller.h"
#include <cmath>
#include <iostream>

using namespace LibOmClient;

namespace OmGtk {

//////////////////// SliderControlGroup ////////////////////////


SliderControlGroup::SliderControlGroup(PortModel* pm)
: ControlGroup(pm),
  m_last_signalled_val(0.0),
  m_enable_signal(false),
  m_name_label(pm->name()),
  m_range_label("Range: "),
  m_slider(pm->user_min(), pm->user_max(), (pm->is_integer() ? 1.0 : 0.0001)),
  m_value_spinner(1.0, (pm->is_integer() ? 0 : 4)),
  m_min_spinner(1.0, (pm->is_integer() ? 0 : 4)), // climb rate, digits
  m_max_spinner(1.0, (pm->is_integer() ? 0 : 4))
{
	m_slider.set_increments(1.0, 10.0);

	// Compensate for crazy plugins
	if (m_port_model->user_max() <= m_port_model->user_min()) {
		m_port_model->user_max(m_port_model->user_min() + 1);
		m_slider.set_range(m_port_model->user_min(), m_port_model->user_max());
	}
	m_slider.property_draw_value() = false;
	
	string name_label = "<span weight=\"bold\">";
	name_label += pm->name() + "</span>";
	m_name_label.set_markup(name_label);

	m_value_spinner.set_range(-99999, 99999);
	m_value_spinner.set_value(m_port_model->value());
	m_value_spinner.set_increments(1.0, 10.0);
	m_value_spinner.signal_value_changed().connect(
		sigc::mem_fun(*this, &SliderControlGroup::update_value_from_spinner));
	m_min_spinner.set_range(-99999, 99999);
	m_min_spinner.set_value(m_port_model->user_min());
	m_min_spinner.set_increments(1.0, 10.0);
	m_min_spinner.signal_value_changed().connect(sigc::mem_fun(*this, &SliderControlGroup::update_range));
	m_max_spinner.set_range(-99999, 99999);
	m_max_spinner.set_value(m_port_model->user_max());
	m_max_spinner.set_increments(1.0, 10.0);
	m_max_spinner.signal_value_changed().connect(sigc::mem_fun(*this, &SliderControlGroup::update_range));

	m_slider_box.pack_start(m_value_spinner, false, false, 1);
	m_slider_box.pack_start(m_slider, true, true, 5);
	m_slider_box.pack_start(m_range_label, false, false, 5);
	m_slider_box.pack_start(m_min_spinner, false, false, 1);
	m_slider_box.pack_start(m_max_spinner, false, false, 1);

	m_slider.set_value(m_port_model->value());

	m_slider.signal_value_changed().connect(
		sigc::mem_fun(*this, &SliderControlGroup::update_value_from_slider));
	
	pack_start(m_name_label);
	pack_start(m_slider_box);

	m_enable_signal = true;

	show_all();
}


void
SliderControlGroup::set_value(float val)
{
	//cerr << "[SliderControlGroup] Setting value to " << val << endl;
	m_enable_signal = false;
	m_slider.set_value(val);
	m_value_spinner.set_value(val);
	m_port_model->value(val);
	m_enable_signal = true;
}


void
SliderControlGroup::set_min(float val)
{
	m_enable_signal = false;
	m_min_spinner.set_value(val);
	m_enable_signal = true;
}


void
SliderControlGroup::set_max(float val)
{
	m_enable_signal = false;
	m_max_spinner.set_value(val);
	m_enable_signal = true;
}


void
SliderControlGroup::update_range()
{
	m_slider.set_range(m_min_spinner.get_value(), m_max_spinner.get_value());
	m_port_model->user_min(m_min_spinner.get_value());
	m_port_model->user_max(m_max_spinner.get_value());
	if (m_enable_signal) {
		char temp_buf[16];
		snprintf(temp_buf, 16, "%f", m_port_model->user_min());
		controller->set_metadata(m_port_model->path(), "user-min", temp_buf);
		snprintf(temp_buf, 16, "%f", m_port_model->user_max());
		controller->set_metadata(m_port_model->path(), "user-max", temp_buf);
	}
}


void
SliderControlGroup::update_value_from_slider()
{
	if (m_enable_signal) {
		float value = m_slider.get_value();
		m_value_spinner.set_value(value);
		if (value != m_last_signalled_val) {
			m_last_signalled_val = value;
			m_signal_value_changed.emit(value);
			m_port_model->value(value);
		}
	}
}


void
SliderControlGroup::update_value_from_spinner()
{
	if (m_enable_signal) {
		float value = m_value_spinner.get_value();
		if (value < m_min_spinner.get_value())
			m_min_spinner.set_value(value);
		if (value > m_max_spinner.get_value())
			m_max_spinner.set_value(value);

		m_enable_signal = false;
		update_range();
		m_slider.set_value(m_value_spinner.get_value());
		m_enable_signal = true;

		if (value != m_last_signalled_val) {
			m_last_signalled_val = value;
			m_signal_value_changed.emit(value);
		}
		m_port_model->value(value);
	}
}


/////////////// IntegerControlGroup ////////////////


IntegerControlGroup::IntegerControlGroup(PortModel* pm)
: ControlGroup(pm),
  m_enable_signal(false),
  m_alignment(0.5, 0.5, 0.0, 0.0),
  m_name_label(pm->name()),
  m_spinner(1.0, 0)
{
	string name_label = "<span weight=\"bold\">";
	name_label += pm->name() + "</span>";
	m_name_label.set_markup(name_label);

	m_spinner.set_range(-99999, 99999);
	m_spinner.set_value(m_port_model->value());
	m_spinner.signal_value_changed().connect(
		sigc::mem_fun(*this, &IntegerControlGroup::update_value));
	m_spinner.set_increments(1, 10);
	
	m_alignment.add(m_spinner);
	pack_start(m_name_label);
	pack_start(m_alignment);

	m_enable_signal = true;

	show_all();
}


void
IntegerControlGroup::set_value(float val)
{
	//cerr << "[IntegerControlGroup] Setting value to " << val << endl;
	m_enable_signal = false;
	m_spinner.set_value(val);
	m_port_model->value(val);
	m_enable_signal = true;
}


void
IntegerControlGroup::update_value()
{
	if (m_enable_signal) {
		float value = m_spinner.get_value();
		m_signal_value_changed.emit(value);
		m_port_model->value(value);
	}
}


/////////////// ToggleControlGroup ////////////////


ToggleControlGroup::ToggleControlGroup(PortModel* pm)
: ControlGroup(pm),
  m_enable_signal(false),
  m_alignment(0.5, 0.5, 0.0, 0.0),
  m_name_label(pm->name())
{
	string name_label = "<span weight=\"bold\">";
	name_label += pm->name() + "</span>";
	m_name_label.set_markup(name_label);

	set_value(m_port_model->value());
	m_checkbutton.signal_toggled().connect(
		sigc::mem_fun(*this, &ToggleControlGroup::update_value));
	
	m_alignment.add(m_checkbutton);
	pack_start(m_name_label);
	pack_start(m_alignment);

	m_enable_signal = true;

	show_all();
}


void
ToggleControlGroup::set_value(float val)
{
	//cerr << "[ToggleControlGroup] Setting value to " << val << endl;
	m_enable_signal = false;
	m_checkbutton.set_active( (val > 0.0f) );
	m_port_model->value(val);
	m_enable_signal = true;
}


void
ToggleControlGroup::update_value()
{
	if (m_enable_signal) {
		float value = m_checkbutton.get_active() ? 1.0f : 0.0f;
		m_signal_value_changed.emit(value);
		m_port_model->value(value);
	}
}


} // namespace OmGtk
